/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include "Ccommon.H"

#include "fatal.H"
#include "erikyyylist.H"
#include <stdlib.h>

#define MODE_init1 200
#define MODE_serverready 201


int Ccommon_ready;
static int commmodus;


// ************************************** Ccommon_init
void Ccommon_init()
{
  Ccommon_ready=0;
  commmodus=MODE_init1;
};


// ************************************** Ccommon_deinit
void Ccommon_deinit()
{
};


// ************************************** Ccommon_input
void Ccommon_input(uint8 msgart,char *message,uint16 len)
{
  switch (commmodus)
    {
    case MODE_init1:
      if (msgart==MSG_ERIKYYYPHONE_SERVERIDENTIFY_1) {
	if (strcmp(message,MSG_ERIKYYYPHONE_SERVERIDENTIFY_1_BLOCK)) 
	  fatalfehler("main(): server failed to identify himself");
	Ccommon_logmsg("Server identified itself. READY");
	commmodus=MODE_serverready;
	Ccommon_output(MSG_ERIKYYYPHONE_IAMCLIENT,NULL,0);
	Ccommon_ready=1;
      }else{
	fatalfehler("main(): server sends wrong message in init mode");
      };
      break;
    case MODE_serverready:
      switch (msgart)
	{
	case MSG_SERVERCONNECTION_BORN:
	  {
	    if (len!=sizeof(serverconnection_statustype))
	      fatalfehler("Ccommon_input: MSG_SERVERCONNECTION_BORN: wrong length of msg");
	    Ccommon_serverconnect_addnew((serverconnection_statustype *)message);
	    break;
	  };
	case MSG_SERVERCONNECTION_DEAD:
	  {
	    if (len!=sizeof(int32))
	      fatalfehler("Ccommon_input: MSG_SERVERCONNECTION_DEAD: wrong length of msg");
	    Ccommon_serverconnect_del(*((int32*)message));
	    break;
	  };
	case MSG_SERVERCONNECTION_MODIFY:
	  {
	    if (len!=sizeof(serverconnection_statustype))
	      fatalfehler("Ccommon_input: MSG_SERVERCONNECTION_MODIFY: wrong length of msg");
	    Ccommon_serverconnect_modify((serverconnection_statustype *)message);
	    break;
	  };

	case MSG_OUTCONNECTION_BORN:
	  {
	    if (len!=sizeof(outconnection_statustype))
	      fatalfehler("Ccommon_input: MSG_OUTCONNECTION_BORN: wrong length of msg");
	    Ccommon_outconnect_addnew((outconnection_statustype *)message);
	    break;
	  };
	case MSG_OUTCONNECTION_DEAD:
	  {
	    if (len!=sizeof(int32))
	      fatalfehler("Ccommon_input: MSG_OUTCONNECTION_DEAD: wrong length of msg");
	    Ccommon_outconnect_del(*((int32*)message));
	    break;
	  };
	case MSG_OUTCONNECTION_MODIFY:
	  {
	    if (len!=sizeof(outconnection_statustype))
	      fatalfehler("Ccommon_input: MSG_OUTCONNECTION_MODIFY: wrong length of msg");
	    Ccommon_outconnect_modify((outconnection_statustype *)message);
	    break;
	  };

	case MSG_STATUS_OF_RECORD:
	  {
	    if (len!=sizeof(int8)) fatalfehler("Ccommon_input: MSG_STATUS_OF_RECORD: wrong length of msg");
	    Ccommon_recordstat( (*message) ? 1 : 0 );
	    break;
	  };
	case MSG_STATUS_OF_RECORDWISH:
	  {
	    if (len!=sizeof(int8)) fatalfehler("Ccommon_input: MSG_STATUS_OF_RECORD: wrong length of msg");
	    Ccommon_recordwishstat( (*message) ? 1 : 0 );
	    break;
	  };
	case MSG_PLEASE_LOG_THIS_MESSAGE:
	  {
	    Ccommon_logmsg(message);
	    break;
	  };
	default:
	  fatalfehler("main(): server send wrong message while i am MODE_serverready");
	};
      break;
    };
};
