/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include "fatal.H"
#include "inconnection.H"
#include "outconnection.H"
#include "serverconnection.H"
#include "clientconnection.H"
#include "sound.H"
#include "erikyyylist.H"
#include "warter.H"
#include "msgarten.H"
#include "recorder.H"
#include "MSG.H"
#include "version.H"

#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <asm/ioctls.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/resource.h>

int main(int argc,char **argv)
{
  int only8bit=0,intobackground=0,willpriority=0;
  char audiodevice[100]="/dev/dsp";

  printf("\n*** ERIKYYYPHONE "erikyyyphoneversion" SERVER ***\n\n");
  printf("Initializing...\n");

  WARTER_init();

  // Terminiersignale
  WARTER_add(WARTER_SIGNAL,SIGHUP);
  WARTER_add(WARTER_SIGNAL,SIGINT);
  WARTER_add(WARTER_SIGNAL,SIGQUIT);
  WARTER_add(WARTER_SIGNAL,SIGABRT);
  WARTER_add(WARTER_SIGNAL,SIGTERM);
  WARTER_add(WARTER_SIGNAL,SIGPWR);

  // Zu ignoierende Signale
  WARTER_add(WARTER_SIGNAL,SIGIO);
  WARTER_add(WARTER_SIGNAL,SIGPIPE);
  WARTER_add(WARTER_SIGNAL,SIGTSTP);
  WARTER_add(WARTER_SIGNAL,SIGTTIN);
  WARTER_add(WARTER_SIGNAL,SIGTTOU);

  clientallowancelist=new liste<int32>;

  {
    printf("Parsing options...\n");
    int opt,optionerror=0,alreadyallowed=0;
    opterr=0;
    while ((opt = getopt(argc, argv, "a:8bd:n")) != EOF) {
      switch (opt)
	{
	case 'd':
	  if (strlen(optarg)<100) {
	    strcpy(audiodevice,optarg);
	    printf("i will use %s as the audio device\n",audiodevice);
	  }else{
	    printf("*** option -d has got a too long parameter");
	    optionerror=1;
	  };
	  break;
	case 'a':
	  int32 ip;
	  if (MSG_string2ip(optarg,&ip)) {
	    char *ips;
	    MSG_ip2string(&ips,ip);
	    printf("Allowing client access for %s\n",ips);
	    free(ips);
	    clientallowancelist->addfront(new int32(ip));
	    alreadyallowed=1;
	  }else{
	    printf("*** You specified an invalid IP in a \"-a\" option\n");
	    optionerror=1;
	  };
	  break;
	case 'b':
	  printf("I will go into background...\n");
	  intobackground=1;
	  break;
	case ':':
	  printf("*** There is a missing parameter for one of the options\n");
	  optionerror=1;
	  break;
	case '8':
	  printf("Forcing 8 Bit audio\n");
	  only8bit=1;
	  break;
	case '?':
	  printf("*** You specified an unknown option, or forgot a parameter to an option\n");
	  optionerror=1;
	  break;
	case 'n':
	  printf("Boosting process priority\n");
	  willpriority=1;
	  break;
	default:
	  optionerror=1;
	};
    };
    if (optind<argc) {
      printf("*** You added some unknown characters after your options\n");
      optionerror=1;
    };
    if (!alreadyallowed) {
      printf("*** You forgot to specify at least one \"-a\" option\n");
      optionerror=1;
    };
    if (optionerror) {
      printf("You made some mistake specifying the options...\n"
	     "\n"
	     "\n"
	     "erikyyyphoneserver -a IP [-a IP ...] [-8] [-b] [-d device] [-n]\n"
	     "\n"
	     "The -a option specifies an IP, that a client is allowed to connect from.\n"
	     "A CLIENT IS THE CONTROL PROGRAMM, NOT ANOTHER CONFERENCE PARTICIPANT !!\n"
	     "SO YOU NORMALLY SHOULD ONLY ALLOW CLIENT ACCESS TO 127.0.0.1 (localhost) !!\n"
	     "\n"
	     "The -8 option forces to use 8 Bit audio. Without -8, the server tries to\n"
	     "  use 16 Bit audio if available. ONLY use this if you got a buggy sound driver\n"
	     "The -b option puts the server into background.\n"
	     "The -d option specifies the audio device. /dev/dsp is the default\n"
	     "The -n option gives the server process more priority\n"
	     "\n"
	     "EXAMPLE:\n"
	     "\n"
	     "erikyyyphoneserver -a 127.0.0.1 -a 10.0.0.9\n"
	     "\n"
	     "This starts an erikyyyphoneserver, and possible clients may connect from\n"
	     "127.0.0.1 and/or 10.0.0.9\n");
      exit(1);
    };
    printf("...option parsing done\n");
  };

  SOUND_init(only8bit,audiodevice);

  RECORDER_init();

  inconnectionlist=new liste<inconnection>;
  clientconnectionlist=new liste<clientconnection>;
  outconnectionlist=new liste<outconnection>;
  serverconnectionlist=new liste<serverconnection>;

  MSG_server *serv=new MSG_server(htons(2900));

  if (willpriority) {
    if (setpriority(PRIO_PROCESS,0,-20)<0)
      fatalfehler("setpriority(...) failed. perhaps you aren't user \"root\" ???\n"
		  "running erikyyyphone as root is only needed, if you use the -n option\n");
  };

  printf("Server is running\n");

  if (intobackground) {
    printf("I am now daemonizing..\n\n");
    int erg=fork();
    if (erg<0) fatalfehlererrno("vfork() failed");
    if (erg) exit(0);

    int f = open("/dev/tty", O_RDWR);
    if (f<0) fatalfehlererrno("open(\"/dev/tty\",O_RDWR) failed");
    if (ioctl(f, TIOCNOTTY, 0) < 0) fatalfehlererrno("TIOCNOTTY failed");
    close(f);

    if (setpgrp()<0) fatalfehlererrno("setpgrp() failed");
    int devnull = open("/dev/null", O_RDWR);
    if (devnull >= 0) {
      dup2(devnull, 0);
      dup2(devnull, 1);
      dup2(devnull, 2);
      FATAL_tosyslog=1;
      if (devnull > 2) close(devnull);
    }
  };

  // Tastatur
  if (!intobackground) WARTER_add(WARTER_READ,0);

  int ende=0;
  do{
    // ***** Event Scheduler
    WARTER_warte();
    printf(".");
    fflush(stdout);

    // ***** Bearbeite neue Connections
    {
      int neu;
      int32 from;
      while (serv->acceptnew(&neu,&from)) inconnectionlist->addfront(new inconnection(neu,from));
    };

    // ***** Abarbeite die inconnections
    {
      liste_elem<inconnection> *akt=inconnectionlist->first;
      while(akt) {
	if (akt->objptr->handleevent()) {
	  akt=akt->next;
	}else{
	  inconnectionlist->del_and_next(&akt);
	};
      };
    };

    // ***** Abarbeite die outconnections
    {
      liste_elem<outconnection> *akt=outconnectionlist->first;
      while(akt) {
	if (akt->objptr->handleevent()) {
	  akt=akt->next;
	}else{
	  outconnectionlist->del_and_next(&akt);
	};
      };
    };

    // ***** Abarbeite die clientconnections
    {
      liste_elem<clientconnection> *akt=clientconnectionlist->first;
      while(akt) {
	if (akt->objptr->handleevent()) {
	  akt=akt->next;
	}else{
	  clientconnectionlist->del_and_next(&akt);
	};
      };
    };

    // ***** Abarbeite die Serverconnections
    {
      liste_elem<serverconnection> *akt=serverconnectionlist->first;
      while(akt) {
	// ***** Digitalisiere  hier, damit wenn moechterecord, dann auch machen.
	RECORDER_handleevent();

	if (akt->objptr->handleevent()) {
	  akt=akt->next;
	}else{
	  serverconnectionlist->del_and_next(&akt);
	};
      };
    };

    // ***** Digitalisiere  hier, damit wenn keine serverconnections da sind, auch alles richtig :)
    RECORDER_handleevent();

    // ***** Bearbeite Eingaben auf STDIN
    if (!intobackground) {
      if (WARTER_handle_aktiv(WARTER_READ,0)) {
	char b;
	read(0,&b,1);
	printf("I got a char. it is <%c>\n",b);
      };
    };

    // ***** Reagiere auf Signale
    if (WARTER_signal_aktiv_und_reset(SIGINT)) {
      printf("I got a SIGINT signal\n");
      ende=1;
    };

    if (WARTER_signal_aktiv_und_reset(SIGTERM)) {
      printf("I got a SIGTERM signal\n");
      ende=1;
    };
    if (WARTER_signal_aktiv_und_reset(SIGHUP)) {
      printf("I got a SIGHUP signal\n");
      ende=1;
    };
    if (WARTER_signal_aktiv_und_reset(SIGQUIT)) {
      printf("I got a SIGQUIT signal\n");
      ende=1;
    };
    if (WARTER_signal_aktiv_und_reset(SIGPWR)) {
      printf("!!! SYSTEM POWER FAILING !!!\n");
      ende=1;
    };
    if (WARTER_signal_aktiv_und_reset(SIGABRT)) {
      printf("I got a SIGABRT signal\n");
      ende=1;
    };
  }while(!ende);
  printf("Shutting down...\n");

  // Tastatur
  if (!intobackground) WARTER_del(WARTER_READ,0);
  
  delete serv;

  delete serverconnectionlist;
  delete outconnectionlist;
  delete clientconnectionlist;
  delete inconnectionlist;
  
  RECORDER_deinit();

  SOUND_deinit();

  delete clientallowancelist;

  // Terminiersignale
  WARTER_del(WARTER_SIGNAL,SIGHUP);
  WARTER_del(WARTER_SIGNAL,SIGINT);
  WARTER_del(WARTER_SIGNAL,SIGQUIT);
  WARTER_del(WARTER_SIGNAL,SIGABRT);
  WARTER_del(WARTER_SIGNAL,SIGTERM);
  WARTER_del(WARTER_SIGNAL,SIGPWR);

  // Zu ignoierende Signale
  WARTER_del(WARTER_SIGNAL,SIGIO);
  WARTER_del(WARTER_SIGNAL,SIGPIPE);
  WARTER_del(WARTER_SIGNAL,SIGTSTP);
  WARTER_del(WARTER_SIGNAL,SIGTTIN);
  WARTER_del(WARTER_SIGNAL,SIGTTOU);

  WARTER_deinit();

  printf("\nByebye. Have a nice day ;)\n\n");
  return(0);
};
