/*
  
  This file is part of ERIKYYYPHONE
  Copyright (C) 1996 Erik Thiele
  see README for details and contact adresses.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include "sound.H"

#include "fatal.H"
extern "C" {
#include "gsmwrapper.h"
};
#include "warter.H"
#include "allgtypedef.H"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>


static gsm gsmhandle;
static gsm_signal *gsmsamples;

static int audioread;
static int audiowrite;

static int habeselectbug;
static int erstesmalrecording;

static int recordingmode;
static int haberecorded;

static int playingmode;
static int habeplayed;

static int recordlock;
static int playlock;

int ANZ_BYTES_PRO_COMPRESSEDGSM;
int ANZ_SAMPLES_PRO_GSM;
int ANZ_SAMPLES_PRO_FRAG;
static int ANZ_BYTES_PRO_SAMPLE;


// ------------------------------------- GSM HANDLING

// ***** SOUND_compressgsm

void SOUND_compressgsm(char *dest, char *src)
{
  for (int u=0;u<ANZ_SAMPLES_PRO_GSM;u++) {
    int16 a=((int16*)src)[u];
    gsmsamples[u]=a;
  };
  
  my_gsm_encode(gsmhandle, gsmsamples+160*0, (gsm_frame)(dest+33*0));
  my_gsm_encode(gsmhandle, gsmsamples+160*1, (gsm_frame)(dest+33*1));
  my_gsm_encode(gsmhandle, gsmsamples+160*2, (gsm_frame)(dest+33*2));
  my_gsm_encode(gsmhandle, gsmsamples+160*3, (gsm_frame)(dest+33*3));
  my_gsm_encode(gsmhandle, gsmsamples+160*4, (gsm_frame)(dest+33*4));
  my_gsm_encode(gsmhandle, gsmsamples+160*5, (gsm_frame)(dest+33*5));
  my_gsm_encode(gsmhandle, gsmsamples+160*6, (gsm_frame)(dest+33*6));
  my_gsm_encode(gsmhandle, gsmsamples+160*7, (gsm_frame)(dest+33*7));
};


// ***** SOUND_uncompressgsm

void SOUND_uncompressgsm(char *dest, char *src)
{
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*0), gsmsamples+160*0)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (0)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*1), gsmsamples+160*1)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (1)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*2), gsmsamples+160*2)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (2)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*3), gsmsamples+160*3)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (3)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*4), gsmsamples+160*4)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (0)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*5), gsmsamples+160*5)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (1)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*6), gsmsamples+160*6)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (2)");
  if (my_gsm_decode(gsmhandle, (gsm_frame)(src+33*7), gsmsamples+160*7)<0)
    fatalfehler("SOUND_uncompressgsm: gsm_decode failed :( (3)");

  for (int u=0;u<ANZ_SAMPLES_PRO_GSM;u++) {
    int16 a=gsmsamples[u];
    ((int16*)dest)[u]=a;
  };
};


// ------------------------------------- INITIALIZING / DEINITIALIZING

// ***** test_bitwidth

static int test_bitwidth(int handle,int w)
{
  int t=w, erg;
  erg=ioctl(handle,SNDCTL_DSP_SETFMT,&t);
  return (!((erg<0) || (t!=w)));
};

// ***** SOUND_init
  
void SOUND_init(int only8bit,char *device)
{  
  // ***** Device oeffnen
  audioread=open(device,O_RDWR);
  if (audioread<0) fatalfehlererrno ("SOUND_init: Unable to open audio device");

  // ***** 16 oder 8 Bit Audio autodetecten, oder 8 Bit force
  if (only8bit) {
    if (!test_bitwidth(audioread,AFMT_U8)) fatalfehler("SOUND_init: Soundcard hasn't got the AFMT_U8 mode");
    ANZ_BYTES_PRO_SAMPLE=1;
  }else{
    if (test_bitwidth(audioread,AFMT_S16_LE)) {
      ANZ_BYTES_PRO_SAMPLE=2;
    }else{
      printf("Hardware cannot do 16 Bit audio...\n");
      if (!test_bitwidth(audioread,AFMT_U8))
	fatalfehler("SOUND_init: Soundcard hasn't got the AFMT_U8 mode");
      ANZ_BYTES_PRO_SAMPLE=1;
    };
  };
  printf("Using %d Bit audio\n",8 << (ANZ_BYTES_PRO_SAMPLE-1));
  
  // ***** Device reopenen
  close(audioread);
  audioread=open("/dev/dsp",O_RDWR);
  if (audioread<0) fatalfehlererrno ("SOUND_init: Unable to open audio device");
  audiowrite=audioread; // wenn != dann folgende inits auch mit audiowrite machen !

  // ***** Fragmentierung einstellen.
  {
    uint t,wunsch=0x000A;
    ANZ_SAMPLES_PRO_FRAG=1<<wunsch;
    if (ANZ_BYTES_PRO_SAMPLE==2) wunsch++;
    t=wunsch+0x00100000;
    if (ioctl(audioread,SNDCTL_DSP_SETFRAGMENT,&t)<0)
      fatalfehlererrno ("SOUND_init: Unable to do SNDCTL_DSP_SETFRAGMENT (audio)");
    if ((t&0xffff)!=wunsch)
      fatalfehler("SOUND_init: Unable to fragmentate the soundcard (audio)");
    //    printf("I am using %d Fragments of Sound, each one is %d bytes big.\n",t>>16,1<<(t&0xffff));
  };

  // ***** Anzahl Bits einstellen
  {
    int t,erg,wunsch;
    wunsch=(ANZ_BYTES_PRO_SAMPLE==1) ? AFMT_U8 : AFMT_S16_LE;
    t=wunsch;
    erg=ioctl(audioread,SNDCTL_DSP_SETFMT,&t);
    if ((erg<0)||(t!=wunsch)) fatalfehler("SOUND_init: this really mustn't happen. buy a new computer :)");
  };

  // ***** Samplingrate einstellen
  {
    int t=8000;
    if (ioctl(audioread,SNDCTL_DSP_SPEED,&t)<0) fatalfehlererrno("SOUND_init: unable to set Sampling rate");
    printf("Working with %d Hz sampling rate.\n",t);
    if (t!=8000) printf("8000 Hz would be better... ;)\n");
  };

  // ***** Selectbug suchen
  {
    /*    timeval timeout={2,0};
    fd_set reader;
    FD_ZERO(&reader);
    FD_SET(audioread,&reader);
    if (select(getdtablesize(),&reader,NULL,NULL,&timeout)<0)
      fatalfehlererrno("SOUND_init: select system call failed.");
    if (!(FD_ISSET(audioread,&reader))) {
      printf("The Sounddriver has got the select-bug. Internal Bugfix activated :-)\n");
      habeselectbug=1;
    }else{
      printf("The Sounddriver hasn't got the select-bug ... WONDERFUL ! :-)\n");
      habeselectbug=0;
    };

    if (ioctl(audioread,SNDCTL_DSP_RESET,0)<0)
      fatalfehlererrno("SOUND_init: RESET failed");*/
    //    printf("I define you have some kind of select() bug, because even if it seems,\n");
    //    printf("that you don't have it, you still have it... (sounddriver problem) :-(\n");
    habeselectbug=1;
  };

  erstesmalrecording=0;
  recordingmode=0;
  playingmode=0;
  haberecorded=0;
  habeplayed=0;
  playlock=0;
  recordlock=0;

  ANZ_BYTES_PRO_COMPRESSEDGSM=33*8;                     // Festgelegt durch GSM Toolkit
  ANZ_SAMPLES_PRO_GSM=160*8;                            // Festgelegt durch GSM Toolkit

  if (!(gsmhandle = my_gsm_create())) fatalfehler("SOUND_init(): GSM_CREATE: error");
  gsmsamples=new gsm_signal[ANZ_SAMPLES_PRO_GSM];

  /*  printf("ANZ_BYTES_PRO_COMPRESSEDGSM %d\n",ANZ_BYTES_PRO_COMPRESSEDGSM);
  printf("ANZ_SAMPLES_PRO_GSM %d\n",ANZ_SAMPLES_PRO_GSM);
  printf("ANZ_SAMPLES_PRO_FRAG %d\n",ANZ_SAMPLES_PRO_FRAG);*/
};


// ***** SOUND_deinit

void SOUND_deinit()
{
  delete [] gsmsamples;
  if (recordingmode) fatalfehler("SOUND_deinit: recording ist noch aktiviert");
  if (playingmode) fatalfehler("SOUND_deinit: playing ist noch aktiviert");
  if (recordlock) fatalfehler("SOUND_deinit: recordlock!=0");
  if (playlock) fatalfehler("SOUND_deinit: playlock!=0");
  if (audiowrite!=audioread) close(audiowrite);
  if (ioctl(audioread,SNDCTL_DSP_RESET,0)<0)
    fatalfehlererrno("SOUND_deinit: RESET failed");
  close(audioread);
  my_gsm_destroy(gsmhandle);
};


// ------------------------------------- RECORDING

// ***** SOUND_setrecordlock

int SOUND_setrecordlock(int neu)
{
  if (neu) {
    if (recordlock) {
      return(0);
    }else{
      if (playlock) return(0);
      recordlock=1;
      return(1);
    };
  }else{
    if (recordlock) {
      recordlock=0;
      WARTER_reenter=1; // ein anderer kanns jetzt kriegen ;)
      return(1);
    }else{
      fatalfehler("SOUND_setrecordlock: soll ausschalten, bin aber schon aus.");
#ifdef kickwarn
      return(0);
#endif
    };
  };
};

// ***** SOUND_setrecordingmode

void SOUND_setrecordingmode(int neumode)
{
  if (!recordlock) fatalfehler("SOUND_setrecordingmode: recordlock=0");
  if (neumode) {
    if (recordingmode) {
      fatalfehler("SOUND_setrecordingmode(1): recording already active");
    }else{
      if (habeplayed) {
	if (ioctl(audiowrite,SNDCTL_DSP_SYNC,0)<0) // DIES kann Zeit brauchen und Blocken :-), sonst nix
	  fatalfehlererrno("SOUND_setrecordingmode: ioctl SYNC failed");
	if (ioctl(audiowrite,SNDCTL_DSP_RESET,0)<0)
	  fatalfehlererrno("SOUND_setrecordingmode: ioctl RESET failed");
	habeplayed=0;
      };
      recordingmode=1;
      if (habeselectbug) {
	erstesmalrecording=1;
	WARTER_reenter=1;
      };
      WARTER_add(WARTER_READ,audioread);
      haberecorded=1;
    };
  }else{
    if (recordingmode) {
      if (ioctl(audioread,SNDCTL_DSP_RESET,0)<0)
	fatalfehlererrno("SOUND_setrecordingmode: ioctl RESET failed (2)");
      recordingmode=0;
      WARTER_del(WARTER_READ,audioread);
      WARTER_reenter=1;  // setplayingmode ist jetzt moeglich
    }else{
      fatalfehler("SOUND_setrecordingmode(0): recording already stopped");
    };
  };
};

// ***** SOUND_recordblock

void SOUND_recordblock(buffer *buf)
{
  if (!recordingmode) fatalfehler("SOUND_recordblock: ich nehm doch garnet auf !!!");
  if (!(SOUND_recordable())) fatalfehler("SOUND_recordblock: called obwohl !SOUND_recordable()");
  if ((buf->bufsize)&1) fatalfehler("SOUND_recordblock: uebergebener Buffer hat ungerade Groesse");

  if (erstesmalrecording) erstesmalrecording=0;

  buf->changesize((buf->bufsize)+(ANZ_SAMPLES_PRO_FRAG*2)); // immer 16 bit :-)
  int erg=read(audioread,(buf->data)+(buf->bufsize)-
	       ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE,
	       ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE);
  if (erg<0) fatalfehlererrno("SOUND_recordblock: Read(...)<0");
  if (erg!=ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE)
    fatalfehler("SOUND_recordblock: read hat nicht alle bytes gereaded :-(");

  if (ANZ_BYTES_PRO_SAMPLE==1) {
    int16 x, *dest=(int16*)((buf->data)+(buf->bufsize)-(ANZ_SAMPLES_PRO_FRAG*2));
    int8 *src=(int8*)((buf->data)+(buf->bufsize)-(ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE));
    for (int i=0;i<ANZ_SAMPLES_PRO_FRAG;i++) {
      x=*src;
      x ^= 0x80;
      *dest=x << 8;
      src++;
      dest++;
    };
  };
};

// ***** SOUND_recordable

int SOUND_recordable()
{
  if (!recordingmode) fatalfehler("SOUND_recordable: ich nehm doch garnet auf !!!");
  if (erstesmalrecording) {
    return(1);
  }else{
    return (WARTER_handle_aktiv(WARTER_READ,audioread));
  };
};


// ------------------------------------- PLAYING

// ***** SOUND_setplaylock

int SOUND_setplaylock(int neu)
{
  if (neu) {
    if (playlock) {
      return(0);
    }else{
      if (recordlock) return(0);
      playlock=1;
      return(1);
    };
  }else{
    if (playlock) {
      playlock=0;
      WARTER_reenter=1; // ein anderer kanns jetzt kriegen :)
      return(1);
    }else{
      fatalfehler("SOUND_setplaylock: soll ausschalten, bin aber schon aus.");
#ifdef kickwarn
      return(0);
#endif
    };
  };
};

// ***** SOUND_setplayingmode

void SOUND_setplayingmode(int neumode)
{
  if (!playlock) fatalfehler("SOUND_setplayingmode: playlock=0");
  if (neumode) {
    if (playingmode) {
      fatalfehler("SOUND_setplayingmode(1): playing already active");
    }else{
      if (haberecorded) {
	if (ioctl(audioread,SNDCTL_DSP_SYNC,0)<0)
	  fatalfehlererrno("SOUND_setplayingmode: ioctl SYNC failed");
	if (ioctl(audioread,SNDCTL_DSP_RESET,0)<0)
	  fatalfehlererrno("SOUND_setplayingmode: ioctl RESET failed");
	haberecorded=0;
      };
      playingmode=1;
      WARTER_add(WARTER_WRITE,audiowrite);
      habeplayed=1;
    };
  }else{
    if (playingmode) {
      playingmode=0;
      WARTER_del(WARTER_WRITE,audiowrite);
      WARTER_reenter=1;  // setrecording wird jetzt moeglich.
    }else{
      fatalfehler("SOUND_setplayingmode(0): playing already stopped");
    };
  };
};

// ***** SOUND_playblock

void SOUND_playblock(buffer *buf)
{
  if (!playingmode) fatalfehler("SOUND_playblock: ich spiel doch garnet ab !!!");
  if (!(SOUND_playable())) fatalfehler("SOUND_playblock: called obwohl !SOUND_playable()");
  if ((buf->bufsize)&1) fatalfehler("SOUND_playblock: uebergebener Buffer hat ungerade Groesse");
  if ((buf->bufsize)<ANZ_SAMPLES_PRO_FRAG*2) fatalfehler("SOUND_playblock: buffer zu klein");

  if (ANZ_BYTES_PRO_SAMPLE==1) {
    int8 *dest=(int8*)(buf->data);
    int16 x,*src=(int16*)(buf->data);
    for (int i=0;i<ANZ_SAMPLES_PRO_FRAG;i++) {
      x=*src;
      x >>= 8;
      *dest = x ^ 0x80;
      src++;
      dest++;
    };
  };

  int erg=write(audiowrite,buf->data,ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE);
  if (erg<0) fatalfehlererrno("SOUND_playblock: write(...)<0");
  if (erg!=ANZ_SAMPLES_PRO_FRAG*ANZ_BYTES_PRO_SAMPLE)
    fatalfehler("SOUND_playblock: write hat nicht alle bytes gewrittened :-(");

  buf->vornekill(ANZ_SAMPLES_PRO_FRAG*2);
};

// ***** SOUND_playable

int SOUND_playable()
{
  if (!playingmode) fatalfehler("SOUND_playblock: ich spiel doch garnet ab !!!");
  return(WARTER_handle_aktiv(WARTER_WRITE,audiowrite));
};
